from functools import cached_property
from traitlets import List, Unicode
from ipyvuetify import VuetifyTemplate
from glue.core import HubListener

from jdaviz.core.template_mixin import DatasetSelect, SelectPluginComponent
from lcviz.events import EphemerisComponentChangedMessage

__all__ = ['EphemerisSelect', 'EphemerisSelectMixin']


class EphemerisSelect(SelectPluginComponent):
    """
    Plugin select for ephemeris components defined by the Ephemeris plugin.

    Useful API methods/attributes:

    * :meth:`~SelectPluginComponent.choices`
    * ``selected``
    * :attr:`selected_obj`
    * :meth:`~SelectPluginComponent.select_default`
    """

    """
    Traitlets (in the object, custom traitlets in the plugin):

    * ``items`` (list of dicts with keys: label)
    * ``selected`` (string)

    Properties (in the object only):

    * ``selected_obj``

    To use in a plugin:

    * create traitlets with default values
    * register with all the automatic logic in the plugin's init by passing the string names
      of the respective traitlets
    * use component in plugin template (see below)
    * refer to properties above based on the interally stored reference to the
      instantiated object of this component

    Example template (label and hint are optional)::

      <plugin-ephemeris-select
        :items="ephemeris_items"
        :selected.sync="ephemeris_selected"
        label="Ephemeris"
        hint="Select ephemeris."
      />

    """
    def __init__(self, plugin, items, selected,
                 default_text='No ephemeris', manual_options=[],
                 default_mode='first'):
        """
        Parameters
        ----------
        plugin
            the parent plugin object
        items : str
            the name of the items traitlet defined in ``plugin``
        selected : str
            the name of the selected traitlet defined in ``plugin``
        default_text : str or None
            the text to show for no selection.  If not provided or None, no entry will be provided
            in the dropdown for no selection.
        manual_options: list
            list of options to provide that are not automatically populated by ephemerides.  If
            ``default`` text is provided but not in ``manual_options`` it will still be included as
            the first item in the list.
        """
        super().__init__(plugin, items=items, selected=selected,
                         default_text=default_text, manual_options=manual_options,
                         default_mode=default_mode)
        self.hub.subscribe(self, EphemerisComponentChangedMessage,
                           handler=self._ephem_component_change)

    @cached_property
    def ephemeris_plugin(self):
        return self.app._jdaviz_helper.plugins.get('Ephemeris', None)

    @cached_property
    def selected_obj(self):
        if self.selected in self._manual_options:
            return None
        return self.ephemeris_plugin.ephemerides.get(self.selected, None)

    def get_data_for_dataset(self, dataset, ycomp='flux'):
        if not isinstance(dataset, DatasetSelect):  # pragma: no cover
            raise ValueError("dataset must be DatasetSelect object")
        if self.selected in self._manual_options:
            return dataset.selected_obj
        return self.ephemeris_plugin.get_data(dataset.selected, self.selected)

    def _ephem_component_change(self, msg=None):
        type = getattr(msg, 'type', None)
        if type == 'remove' and msg.old_lbl in self.choices:
            self.items = [item for item in self.items if item['label'] != msg.old_lbl]
            self._apply_default_selection()
        elif type == 'rename' and msg.old_lbl in self.choices:
            was_selected = self.selected == msg.old_lbl
            self.items = [item if item['label'] != msg.old_lbl else {'label': msg.new_lbl}
                          for item in self.items]
            if was_selected:
                self.selected = msg.new_lbl
        elif type == 'add' and msg.new_lbl not in self.choices:
            self.items = self.items + [{'label': msg.new_lbl}]
        else:
            # something might be out of sync, build from scratch
            manual_items = [{'label': label} for label in self.manual_options]
            self.items = manual_items + [{'label': component}
                                         for component in self.ephemeris_plugin.ephemerides.keys()]
            self._apply_default_selection()


class EphemerisSelectMixin(VuetifyTemplate, HubListener):
    """
    Applies the EphemerisSelect component as a mixin in the base plugin.  This
    automatically adds traitlets as well as new properties to the plugin with minimal
    extra code.  For multiple instances or custom traitlet names/defaults, use the
    component instead.

    To use in a plugin:

    * add ``EphemerisSelectMixin`` as a mixin to the class
    * use the traitlets available from the plugin or properties/methods available from
      ``plugin.ephemeris``.

    Example template (label and hint are optional)::

      <plugin-ephemeris-select
        :items="ephemeris_items"
        :selected.sync="ephemeris_selected"
        label="Ephemeris"
        hint="Select ephemeris."
      />

    """
    ephemeris_items = List().tag(sync=True)
    ephemeris_selected = Unicode().tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ephemeris = EphemerisSelect(self, 'ephemeris_items', 'ephemeris_selected')



import numpy as np
from time import time

from traitlets import Bool, Float, Unicode, observe

from jdaviz.core.custom_traitlets import FloatHandleEmpty, IntHandleEmpty
from jdaviz.core.events import ViewerAddedMessage
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin,
                                        DatasetSelectMixin, AddResultsMixin,
                                        skip_if_no_updates_since_last_active,
                                        with_spinner)
from jdaviz.core.user_api import PluginUserApi

from lcviz.marks import LivePreviewTrend, LivePreviewFlattened
from lcviz.utils import data_not_folded
from lcviz.viewers import TimeScatterView, PhaseScatterView
from lcviz.parsers import _data_with_reftime

__all__ = ['Flatten']


@tray_registry('flatten', label="Flatten")
class Flatten(PluginTemplateMixin, DatasetSelectMixin, AddResultsMixin):
    """
    See the :ref:`Flatten Plugin Documentation <flatten>` for more details.

    Only the following attributes and methods are available through the
    public plugin API.

    * ``show_live_preview`` : bool
        Whether to show the live-preview of the (unnormalized) flattened light curve
    * ``show_trend_preview`` : bool
        Whether to show the live-preview of the trend curve used to flatten the light curve
    * ``default_to_overwrite``
    * ``dataset`` (:class:`~jdaviz.core.template_mixin.DatasetSelect`):
      Dataset to flatten.
    * ``add_results`` (:class:`~jdaviz.core.template_mixin.AddResults`)
    * ``window_length``
    * ``polyorder``
    * ``break_tolerance``
    * ``niters``
    * ``sigma``
    * ``unnormalize``
    * :meth:`flatten`
    """
    template_file = __file__, "flatten.vue"
    uses_active_status = Bool(True).tag(sync=True)

    show_live_preview = Bool(True).tag(sync=True)
    show_trend_preview = Bool(True).tag(sync=True)
    default_to_overwrite = Bool(True).tag(sync=True)
    flatten_err = Unicode().tag(sync=True)

    window_length = IntHandleEmpty(101).tag(sync=True)
    polyorder = IntHandleEmpty(2).tag(sync=True)
    break_tolerance = IntHandleEmpty(5).tag(sync=True)
    niters = IntHandleEmpty(3).tag(sync=True)
    sigma = FloatHandleEmpty(3).tag(sync=True)
    unnormalize = Bool(False).tag(sync=True)

    last_live_time = Float(0).tag(sync=True)
    previews_temp_disable = Bool(False).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # do not support flattening data in phase-space
        self.dataset.add_filter(data_not_folded)

        # marks do not exist for the new viewer, so force another update to compute and draw
        # those marks
        self.hub.subscribe(self, ViewerAddedMessage, handler=lambda _: self._live_update())

    @property
    def user_api(self):
        expose = ['show_live_preview', 'show_trend_preview', 'default_to_overwrite',
                  'dataset', 'add_results',
                  'window_length', 'polyorder', 'break_tolerance',
                  'niters', 'sigma', 'unnormalize', 'flatten']
        return PluginUserApi(self, expose=expose)

    @property
    def marks(self):
        trend_marks = {}
        flattened_marks = {}

        for id, viewer in self.app._viewer_store.items():
            needs_trend = isinstance(viewer, TimeScatterView) and not isinstance(viewer, PhaseScatterView)  # noqa
            needs_flattened = isinstance(viewer, (TimeScatterView, PhaseScatterView))
            for mark in viewer.figure.marks:
                if isinstance(mark, LivePreviewTrend):
                    trend_marks[id] = mark
                    needs_trend = False
                elif isinstance(mark, LivePreviewFlattened):
                    flattened_marks[id] = mark
                    needs_flattened = False
                if not needs_trend and not needs_flattened:
                    break
            if needs_trend:
                mark = LivePreviewTrend(viewer, visible=self.is_active)
                viewer.figure.marks = viewer.figure.marks + [mark]
                trend_marks[id] = mark
            if needs_flattened:
                mark = LivePreviewFlattened(viewer, visible=self.is_active)
                viewer.figure.marks = viewer.figure.marks + [mark]
                flattened_marks[id] = mark

        return trend_marks, flattened_marks

    @observe('default_to_overwrite', 'dataset_selected')
    def _set_default_results_label(self, event={}):
        '''Generate a label and set the results field to that value'''
        if not hasattr(self, 'dataset'):  # pragma: no cover
            return

        self.add_results.label_whitelist_overwrite = [self.dataset_selected]

        if self.default_to_overwrite:
            self.results_label_default = self.dataset_selected
        else:
            self.results_label_default = f"{self.dataset_selected} (flattened)"

    @with_spinner()
    def flatten(self, add_data=True):
        """
        Flatten the input light curve (``dataset``) using lightkurve.flatten.

        Parameters
        ----------
        add_data : bool
            Whether to add the resulting trace to the application, according to the options
            defined in the plugin.

        Returns
        -------
        output_lc : `~lightkurve.LightCurve`
            The flattened light curve.
        trend_lc : `~lightkurve.LightCurve`
            The trend used to flatten the light curve.
        """
        input_lc = self.dataset.selected_obj
        if input_lc is None:  # pragma: no cover
            raise ValueError("no input dataset selected")

        output_lc, trend_lc = input_lc.flatten(return_trend=True,
                                               window_length=self.window_length,
                                               polyorder=self.polyorder,
                                               break_tolerance=self.break_tolerance,
                                               niters=self.niters,
                                               sigma=self.sigma)

        if self.unnormalize:
            factor = np.nanmedian(trend_lc.flux.value)
            output_lc.flux *= factor
            output_lc.flux_err *= factor
            output_lc.meta['NORMALIZED'] = False

        if add_data:
            # add data to the collection/viewer
            data = _data_with_reftime(self.app, output_lc)
            self.add_results.add_results_from_plugin(data)

        return output_lc, trend_lc

    def _clear_marks(self):
        for mark_set in self.marks:
            for mark in mark_set.values():
                if mark.visible:
                    mark.clear()
                    mark.visible = False

    @observe('is_active', 'show_live_preview', 'show_trend_preview')
    def _toggle_marks(self, event={}):
        live_visible = self.show_live_preview and self.is_active
        trend_visible = self.show_trend_preview and self.is_active

        trend_marks, flattened_marks = self.marks
        for mark in trend_marks.values():
            mark.visible = trend_visible
        for mark in flattened_marks.values():
            mark.visible = live_visible

        if ((live_visible or trend_visible) and
                event.get('name') in ('is_active', 'show_live_preview', 'show_trend_preview')):
            # then the marks themselves need to be updated
            self._live_update(event)

    @observe('dataset_selected',
             'window_length', 'polyorder', 'break_tolerance',
             'niters', 'sigma', 'previews_temp_disable')
    @skip_if_no_updates_since_last_active()
    def _live_update(self, event={}):
        if self.previews_temp_disable:
            return

        start = time()
        try:
            output_lc, trend_lc = self.flatten(add_data=False)
        except Exception as e:
            self.flatten_err = str(e)
            self._clear_marks()
            return
        self.flatten_err = ''

        if event.get('name') not in ('is_active', 'show_live_preview', 'show_trend_preview'):
            # mark visibility hasn't been handled yet
            self._toggle_marks(event)

        if self.unnormalize:
            output_flux = output_lc.flux.value
        else:
            output_flux = output_lc.flux.value * np.nanmedian(trend_lc.flux.value)

        ref_time = trend_lc.meta.get('reference_time', 0)
        times = trend_lc.time - ref_time
        trend_marks, flattened_marks = self.marks
        for mark in trend_marks.values():
            # TODO: need to account for phasing
            mark.update_ty(times.value, trend_lc.flux.value)
        for mark in flattened_marks.values():
            mark.update_ty(times.value, output_flux)

        self.last_live_time = np.round(time() - start, 2)
        if self.last_live_time > 0.3:
            self.previews_temp_disable = True

    def vue_apply(self, *args, **kwargs):
        try:
            self.flatten(add_data=True)
        except Exception as e:
            self.flatten_err = str(e)
        else:
            self.flatten_err = ''
        if self.add_results.label_overwrite:
            # then this will change the input data without triggering a
            # change to dataset_selected
            self._live_update()



from jdaviz.configs.default.plugins import SubsetPlugin
from jdaviz.core.registries import tray_registry

__all__ = ['SubsetPlugin']


@tray_registry('lcviz-subset-plugin', label="Subset Tools")
class SubsetPlugin(SubsetPlugin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.docs_link = f"https://lcviz.readthedocs.io/en/{self.vdocs}/plugins.html#subset-tools"
        self.can_freeze = True



from jdaviz.core.tools import SidebarShortcutPlotOptions, SidebarShortcutExportPlot

# point to the lcviz-version of plot options instead of jdaviz's
SidebarShortcutPlotOptions.plugin_name = 'lcviz-plot-options'
SidebarShortcutExportPlot.plugin_name = 'lcviz-export-plot'



import astropy.units as u
import ipyvue
import os

from lightkurve import LightCurve

from glue.core.component_id import ComponentID
from glue.core.link_helpers import LinkSame
from jdaviz.core.helpers import ConfigHelper

__all__ = ['LCviz']

_default_time_viewer_reference_name = 'flux-vs-time'

custom_components = {'plugin-ephemeris-select': 'components/plugin_ephemeris_select.vue'}

# Register pure vue component. This allows us to do recursive component instantiation only in the
# vue component file
for name, path in custom_components.items():
    ipyvue.register_component_from_file(None, name,
                                        os.path.join(os.path.dirname(__file__), path))


def _get_range_subset_bounds(self, subset_state, *args, **kwargs):
    # Instead of overriding the jdaviz version of this method on jdaviz.Application,
    # we could put in jdaviz by (1) checking if helper has a
    # _default_time_viewer_reference_name, (2) using the LCviz version if so, and (3)
    # using the jdaviz version otherwise.
    viewer = self.get_viewer(self._jdaviz_helper._default_time_viewer_reference_name)
    light_curve = viewer.data()[0]
    reference_time = light_curve.meta['reference_time']
    if viewer:
        # TODO: use display units once implemented in Glue for ScatterViewer
        # units = u.Unit(viewer.state.x_display_unit)
        units = u.Unit(viewer.time_unit)
    else:  # pragma: no cover
        raise ValueError("Unable to find time axis units")

    region = reference_time + u.Quantity([subset_state.lo * units, subset_state.hi * units])

    return [{"name": subset_state.__class__.__name__,
             "glue_state": subset_state.__class__.__name__,
             "region": region,
             "subset_state": subset_state}]


def _link_new_data(app, reference_data=None, data_to_be_linked=None):
    dc = app.data_collection
    ref_data = dc[reference_data] if reference_data else dc[0]
    linked_data = dc[data_to_be_linked] if data_to_be_linked else dc[-1]

    new_link = LinkSame(
        cid1=ref_data.world_component_ids[0],
        cid2=linked_data.world_component_ids[0],
        data1=ref_data,
        data2=linked_data,
        labels1=ref_data.label,
        labels2=linked_data.label
    )

    dc.add_link(new_link)
    return


class LCviz(ConfigHelper):
    _default_configuration = {
        'settings': {'configuration': 'lcviz',
                     'visible': {'menu_bar': False,
                                 'toolbar': True,
                                 'tray': True,
                                 'tab_headers': True},
                     'dense_toolbar': False,
                     'context': {'notebook': {'max_height': '600px'}}},
        'toolbar': ['g-data-tools', 'g-subset-tools', 'lcviz-coords-info'],
        'tray': ['lcviz-metadata-viewer', 'lcviz-plot-options', 'lcviz-subset-plugin',
                 'lcviz-markers', 'flatten', 'frequency-analysis', 'ephemeris',
                 'binning', 'lcviz-export-plot'],
        'viewer_area': [{'container': 'col',
                         'children': [{'container': 'row',
                                       'viewers': [{'name': 'flux-vs-time',
                                                    'plot': 'lcviz-time-viewer',
                                                    'reference': 'flux-vs-time'}]}]}]}

    _component_ids = {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._default_time_viewer_reference_name = _default_time_viewer_reference_name

        # override jdaviz behavior to support temporal subsets
        self.app._get_range_subset_bounds = (
            lambda *args, **kwargs: _get_range_subset_bounds(self.app, *args, **kwargs)
        )

        self.app._link_new_data = (
            lambda *args, **kwargs: _link_new_data(self.app, *args, **kwargs)
        )

        # inject custom css from lcviz_style.vue (on top of jdaviz styles)
        if hasattr(self.app, '_add_style'):
            # will be guaranteed after jdaviz 3.9
            self.app._add_style((__file__, 'lcviz_style.vue'))
        else:
            self.app.set_style_template_file((__file__, 'lcviz_style.vue'))

        # set the link to read the docs
        self.app.docs_link = "https://lcviz.readthedocs.io"

    def load_data(self, data, data_label=None):
        """
        Load data into LCviz.

        Parameters
        ----------
        data : obj or str
            File name or object to be loaded. Supported formats include:

            * ``'filename.fits'`` (or any extension that ``astropy.io.fits``
              supports)
            * `~lightkurve.LightCurve` (extracts the default flux column)
        data_label : str or `None`
            Data label to go with the given data. If not given, this is
            automatically determined from filename or randomly generated.
            The final label shown in LCviz may have additional information
            appended for clarity.
        """
        super().load_data(
            data=data,
            parser_reference='light_curve_parser',
            data_label=data_label
        )

    def get_data(self, data_label=None, cls=LightCurve, subset=None):
        """
        Returns data with name equal to data_label of type cls with subsets applied from
        subset_to_apply.

        Parameters
        ----------
        data_label : str, optional
            Provide a label to retrieve a specific data set from data_collection.
        cls : light curve class, optional
            The type that data will be returned as.
        subset : str, optional
            Subset that is to be applied (as a mask) to the data before it is returned.

        Returns
        -------
        data : cls
            Data is returned as type cls with subsets applied.
        """
        return super()._get_data(data_label=data_label, mask_subset=subset, cls=cls)

    def _phase_comp_lbl(self, component):
        return f'phase:{component}'

    def _set_data_component(self, data, component_label, values):
        if component_label not in self._component_ids:
            self._component_ids[component_label] = ComponentID(component_label)

        if self._component_ids[component_label] in data.components:
            data.update_components({self._component_ids[component_label]: values})
        else:
            data.add_component(values, self._component_ids[component_label])

        data.add_component(values, self._component_ids[component_label])



from glue.core.message import Message

__all__ = ['EphemerisComponentChangedMessage',
           'EphemerisChangedMessage']


class EphemerisComponentChangedMessage(Message):
    """Message emitted when an ephemeris component is added/renamed/removed in the
    ephemeris plugin"""
    def __init__(self, old_lbl, new_lbl, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.old_lbl = old_lbl
        self.new_lbl = new_lbl
        if old_lbl is not None and new_lbl is not None:
            self.type = 'rename'
        elif old_lbl is None and new_lbl is not None:
            self.type = 'add'
        elif old_lbl is not None and new_lbl is None:
            self.type = 'remove'
        else:  # pragma: no cover
            raise ValueError("must provide at least one of old_lbl or new_lbl")


class EphemerisChangedMessage(Message):
    """Message emitted when the parameters of an ephemeris are updated/changed
    in the ephemeris plugin"""
    def __init__(self, ephemeris_label, *args, **kwargs):
        self.ephemeris_label = ephemeris_label




def test_docs_snippets(helper, light_curve_like_kepler_quarter):
    lcviz, lc = helper, light_curve_like_kepler_quarter

    lcviz.load_data(lc)
    # lcviz.show()

    ephem = lcviz.plugins['Ephemeris']
    ephem.period = 4.88780258
    ephem.t0 = 2.43
    ephem.rename_component('default', 'my component name')


def test_plugin_ephemeris(helper, light_curve_like_kepler_quarter):
    helper.load_data(light_curve_like_kepler_quarter)
    ephem = helper.plugins['Ephemeris']

    assert len(helper.app.get_viewer_ids()) == 1
    assert not ephem._obj.phase_viewer_exists

    # setting any value in the ephemeris should create a new viewer
    ephem.period = 3.14
    assert len(helper.app.get_viewer_ids()) == 2
    assert ephem._obj.phase_viewer_exists
    assert ephem._obj.phase_viewer_id in helper.app.get_viewer_ids()

    ephem.t0 = 5
    ephem._obj.vue_period_double()
    assert ephem.period == 3.14 * 2
    ephem._obj.vue_period_halve()
    assert ephem.period == 3.14

    pv = ephem._obj.phase_viewer
    # original limits are set to 0->1 (technically 1-phase_wrap -> phase_wrap)
    assert (pv.state.x_min, pv.state.x_max) == (0.0, 1.0)
    ephem.wrap_at = 0.5
    assert (pv.state.x_min, pv.state.x_max) == (-0.5, 0.5)

    ephem.add_component('custom component')
    assert not ephem._obj.phase_viewer_exists
    ephem.create_phase_viewer()
    assert len(helper.app.get_viewer_ids()) == 3
    assert len(ephem.ephemerides) == 2
    assert 'custom component' in ephem.ephemerides

    ephem.rename_component('custom component', 'renamed custom component')
    assert len(ephem.ephemerides) == 2
    assert 'renamed custom component' in ephem.ephemerides
    assert len(helper.app.get_viewer_ids()) == 3

    assert ephem.component == 'renamed custom component'
    assert ephem.period == 3.14
    assert ephem.ephemeris['period'] == 3.14
    # modify the ephemeris of the NON-selected ephemeris component
    ephem.update_ephemeris(ephem_component='default', period=2)
    assert ephem.period == 3.14
    assert ephem.ephemerides['default']['period'] == 2

    ephem.remove_component('renamed custom component')
    assert len(ephem.ephemerides) == 1
    assert len(helper.app.get_viewer_ids()) == 2
    assert ephem.component == 'default'
    assert ephem.period == 2

    assert ephem.method.selected == 'Lomb-Scargle'
    ephem.method = 'Box Least Squares'
    assert ephem._obj.method_err == ''
    ephem._obj.vue_adopt_period_at_max_power()
    assert ephem.period != 2

    # test that non-zero dpdt does not crash
    ephem.dpdt = 0.005



import os
from glue.config import data_translator
from jdaviz.core.registries import data_parser_registry
import lightkurve

__all__ = ["light_curve_parser"]


@data_parser_registry("light_curve_parser")
def light_curve_parser(app, file_obj, data_label=None, show_in_viewer=True, **kwargs):
    time_viewer_reference_name = app._jdaviz_helper._default_time_viewer_reference_name

    # load local FITS file from disk by its path:
    if isinstance(file_obj, str) and os.path.exists(file_obj):
        if data_label is None:
            data_label = os.path.splitext(os.path.basename(file_obj))[0]

        # read the light curve:
        light_curve = lightkurve.read(file_obj)

    # load a LightCurve object:
    elif isinstance(file_obj, lightkurve.LightCurve):
        light_curve = file_obj

    # make a data label:
    if data_label is not None:
        new_data_label = f'{data_label}'
    else:
        new_data_label = light_curve.meta.get('OBJECT', 'Light curve')
    flux_origin = light_curve.meta.get('FLUX_ORIGIN', None)  # i.e. PDCSAP or SAP
    if flux_origin is not None:
        new_data_label += f'[{flux_origin}]'

    data = _data_with_reftime(app, light_curve)
    app.add_data(data, new_data_label)

    if show_in_viewer:
        app.add_data_to_viewer(time_viewer_reference_name, new_data_label)

        # add to any known phase viewers
        ephem_plugin = app._jdaviz_helper.plugins.get('Ephemeris', None)
        if ephem_plugin is not None:
            for viewer_id in ephem_plugin._obj.phase_viewer_ids:
                app.add_data_to_viewer(viewer_id, new_data_label)


def _data_with_reftime(app, light_curve):
    # grab the first-found reference time in the data collection:
    ff_reference_time = None
    for existing_data in app.data_collection:
        if hasattr(existing_data, 'meta') and 'reference_time' in existing_data.meta:
            ff_reference_time = existing_data.meta.get('reference_time', None)
            if ff_reference_time is not None:
                break

    # convert to glue Data manually, so we may edit the `dt` component if necessary:
    handler, _ = data_translator.get_handler_for(light_curve)
    return handler.to_data(light_curve, reference_time=ff_reference_time)




def test_reset_limits(helper, light_curve_like_kepler_quarter):
    helper.load_data(light_curve_like_kepler_quarter)
    tv = helper.app.get_viewer(helper._default_time_viewer_reference_name)

    orig_xlims = (tv.state.x_min, tv.state.x_max)
    orig_ylims = (tv.state.y_min, tv.state.y_max)
    # set xmin and ymin to midpoints
    new_xmin = (tv.state.x_min + tv.state.x_max) / 2
    new_ymin = (tv.state.y_min + tv.state.y_max) / 2
    tv.state.x_min = new_xmin
    tv.state.y_min = new_ymin

    tv.state._reset_x_limits()
    assert tv.state.x_min == orig_xlims[0]
    assert tv.state.y_min == new_ymin

    tv.state._reset_y_limits()
    assert tv.state.y_min == orig_ylims[0]



from jdaviz.configs.default.plugins import PlotOptions
from jdaviz.core.registries import tray_registry

__all__ = ['PlotOptions']


@tray_registry('lcviz-plot-options', label="Plot Options")
class PlotOptions(PlotOptions):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.docs_link = f"https://lcviz.readthedocs.io/en/{self.vdocs}/plugins.html#plot-options"

    @property
    def user_api(self):
        api = super().user_api
        api._expose += ['marker_visible', 'marker_fill', 'marker_opacity',
                        'marker_size_mode', 'marker_size', 'marker_size_scale',
                        'marker_size_col', 'marker_size_vmin', 'marker_size_vmax',
                        'marker_color_mode', 'marker_color', 'marker_color_col',
                        'marker_colormap', 'marker_colormap_vmin', 'marker_colormap_vmax',
                        'line_visible', 'line_width']

        return api



from jdaviz.configs.default.plugins import ExportViewer
from jdaviz.core.registries import tray_registry

__all__ = ['ExportViewer']


@tray_registry('lcviz-export-plot', label="Export Plot")
class ExportViewer(ExportViewer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.docs_link = f"https://lcviz.readthedocs.io/en/{self.vdocs}/plugins.html#export-plot"



from .binning.binning import *  # noqa
from .coords_info.coords_info import *  # noqa
from .ephemeris.ephemeris import *  # noqa
from .export_plot.export_plot import *  # noqa
from .flatten.flatten import *  # noqa
from .frequency_analysis.frequency_analysis import *  # noqa
from .markers.markers import *  # noqa
from .metadata_viewer.metadata_viewer import *  # noqa
from .plot_options.plot_options import *  # noqa
from .subset_plugin.subset_plugin import *  # noqa



from glue.config import data_translator
from glue.core import Data, Subset
from ipyvue import watch

import os
from glue.core.coordinates import Coordinates
from glue.core.component_id import ComponentID
import numpy as np
from scipy.interpolate import interp1d
from astropy import units as u
from astropy.table import QTable
from astropy.time import Time

from lightkurve import (
    LightCurve, KeplerLightCurve, TessLightCurve, FoldedLightCurve
)

__all__ = ['TimeCoordinates', 'LightCurveHandler', 'data_not_folded']


class TimeCoordinates(Coordinates):
    """
    This is a sub-class of Coordinates that is intended for a time axis
    given by a :class:`~astropy.time.Time` array.
    """

    def __init__(self, times, reference_time=None, unit=u.d):
        if not isinstance(times, Time):  # pragma: no cover
            raise TypeError('values should be a Time instance')
        self._index = np.arange(len(times))
        self._times = times
        self.unit = unit

        # convert to relative time units
        self.reference_time = reference_time
        if self.reference_time is None:
            self.reference_time = times[0]
        self._values = (times - self.reference_time).to(unit)

        super().__init__(n_dim=1)

    @property
    def time_axis(self):
        return self._times

    def world_to_pixel_values(self, *world):
        if len(world) > 1:  # pragma: no cover
            raise ValueError('TimeCoordinates is a 1-d coordinate class '
                             'and only accepts a single scalar or array to convert')
        return interp1d(
            self._values.value, self._index, fill_value='extrapolate'
        )(world[0])

    def pixel_to_world_values(self, *pixel):
        if len(pixel) > 1:  # pragma: no cover
            raise ValueError('SpectralCoordinates is a 1-d coordinate class '
                             'and only accepts a single scalar or array to convert')
        return interp1d(
            self._index, self._values.value, fill_value='extrapolate'
        )(pixel[0])


__all__ = ['LightCurveHandler', 'enable_hot_reloading']


@data_translator(LightCurve)
class LightCurveHandler:
    lc_component_ids = {}

    def to_data(self, obj, reference_time=None):
        is_folded = isinstance(obj, FoldedLightCurve)
        time = obj.time_original if is_folded and hasattr(obj, 'time_original') else obj.time
        time_coord = TimeCoordinates(time, reference_time=reference_time)
        data = Data(coords=time_coord)

        if hasattr(obj, 'label'):
            data.label = obj.label

        data.meta.update(obj.meta)
        data.meta.update(
            {"reference_time": time_coord.reference_time}
        )
        data['dt'] = (obj.time - time_coord.reference_time).to(time_coord.unit)
        data.get_component('dt').units = str(time_coord.unit)

        # LightCurve is a subclass of astropy TimeSeries, so
        # collect all other columns in the TimeSeries:
        for component_label in obj.colnames:

            component_data = getattr(obj, component_label)
            if is_folded and component_label == 'time':
                ephem_comp = obj.meta.get('_LCVIZ_EPHEMERIS', {}).get('ephemeris')
                if ephem_comp is None:
                    continue
                component_label = f'phase:{ephem_comp}'

            if component_label not in self.lc_component_ids:
                self.lc_component_ids[component_label] = ComponentID(component_label)
            cid = self.lc_component_ids[component_label]

            data[cid] = component_data
            if hasattr(component_data, 'unit'):
                try:
                    data.get_component(cid).units = str(component_data.unit)
                except KeyError:  # pragma: no cover
                    continue

        data.meta.update({'uncertainty_type': 'std'})

        # if the anticipated x and y axes are the first two components in the
        # Data object, the viewer will load those components correctly before
        # you hit the call to `viewer.set_plot_axes`:
        reordered_components = {comp.label: comp for comp in data.components}
        dt_comp = reordered_components.pop('dt')
        flux_comp = reordered_components.pop('flux')
        data.reorder_components(
            [dt_comp, flux_comp] +
            list(reordered_components.values())
        )

        return data

    def to_object(self, data_or_subset):
        """
        Convert a glue Data object to a lightkurve.LightCurve object.

        Parameters
        ----------
        data_or_subset : `glue.core.data.Data` or `glue.core.subset.Subset`
            The data to convert to a LightCurve object
        attribute : `glue.core.component_id.ComponentID`
            The attribute to use for the LightCurve data
        """

        if isinstance(data_or_subset, Subset):
            data = data_or_subset.data
            subset_state = data_or_subset.subset_state
        else:
            data = data_or_subset
            subset_state = None

        # Copy over metadata
        kwargs = {'meta': data.meta.copy()}

        # extract a Time object out of the TimeCoordinates object:
        time = data.coords.time_axis

        if subset_state is None:
            # pass through mask of all True's if no glue subset is chosen
            glue_mask = np.ones(len(time)).astype(bool)
        else:
            # get the subset mask from glue:
            glue_mask = data.get_mask(subset_state=subset_state)
            # apply the subset mask to the time array:
            time = time[glue_mask]

        columns = [time]
        names = ['time']

        component_ids = data.main_components

        # we already handled time separately above, and `dt` is only used internally
        # in LCviz, so let's skip those IDs below:
        skip_components = [id for id in component_ids if id.label in ['time', 'dt']]
        for skip_comp in skip_components:
            component_ids.remove(skip_comp)

        for component_id in component_ids:
            if component_id.label in names:
                # avoid duplicate column
                continue
            component = data.get_component(component_id)

            values = component.data[glue_mask]

            if len(values) and isinstance(values[0], Time):
                values = Time(values.base)
            elif hasattr(component, 'units') and component.units != "None":
                values = u.Quantity(values, component.units)

            if component_id.label not in names:
                columns.append(values)
                names.append(component_id.label)

        table = QTable(columns, names=names, masked=True, copy=False)
        return LightCurve(table, **kwargs)


def enable_hot_reloading(watch_jdaviz=True):
    """
    Use ``watchdog`` to perform hot reloading.

    Parameters
    ----------
    watch_jdaviz : bool
        Whether to also watch changes to jdaviz upstream.
    """
    try:
        watch(os.path.dirname(__file__))
        if watch_jdaviz:
            import jdaviz
            watch(os.path.dirname(jdaviz.__file__))
    except ModuleNotFoundError:
        print((
            'Watchdog module, needed for hot reloading, not found.'
            ' Please install with `pip install watchdog`'))


@data_translator(KeplerLightCurve)
class KeplerLightCurveHandler(LightCurveHandler):
    # Works the same as LightCurve
    pass


@data_translator(TessLightCurve)
class TessLightCurveHandler(LightCurveHandler):
    # Works the same as LightCurve
    pass


# plugin component filters
def data_not_folded(data):
    return data.meta.get('_LCVIZ_EPHEMERIS', None) is None



import numpy as np

from jdaviz.configs.default.plugins import Markers
from jdaviz.core.registries import tray_registry

__all__ = ['Markers']


@tray_registry('lcviz-markers', label="Markers")
class Markers(Markers):
    _default_table_values = {'time': np.nan,
                             'phase': np.nan,
                             'ephemeris': '',
                             'flux': np.nan}

    def __init__(self, *args, **kwargs):
        kwargs['headers'] = ['time', 'phase', 'ephemeris', 'flux', 'viewer']
        super().__init__(*args, **kwargs)
        self.docs_link = f"https://lcviz.readthedocs.io/en/{self.vdocs}/plugins.html#markers"

    @property
    def coords_info(self):
        return self.app.session.application._tools['lcviz-coords-info']



from jdaviz.configs.default.plugins import MetadataViewer
from jdaviz.core.registries import tray_registry

__all__ = ['MetadataViewer']


@tray_registry('lcviz-metadata-viewer', label="Metadata")
class MetadataViewer(MetadataViewer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.docs_link = f"https://lcviz.readthedocs.io/en/{self.vdocs}/plugins.html#metadata-viewer"  # noqa



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
from astropy.tests.runner import TestRunner

try:
    from .version import version as __version__
except ImportError:
    __version__ = ''

from . import utils  # noqa
from .utils import enable_hot_reloading  # noqa: F401
from . import state  # noqa

# Expose subpackage API at package level.
from .plugins import *  # noqa
from .parsers import *  # noqa
from .tools import *  # noqa
from .viewers import *  # noqa
from .helper import *  # noqa

# Create the test function for self test
test = TestRunner.make_test_runner_in(os.path.dirname(__file__))
